###PACKAGES###
require(dplyr)
require(tibble)
require(tidyr)
require(readr)
require(stringr)
require(textstem)
require(lexicon)
require(quanteda)
require(quanteda.textmodels)
require(purrr)
require(ggplot2)

###FILEPATHS AND PARAMETERS###

#Filepaths
InPath <- "C:/Users/UserName/Documents/SVD_Illustrative_Example/"
OutPath <- "C:/Users/UserName/Documents/SVD_Illustrative_Example/Results/"

#Dictionary parameters
token_split <- " " 
lemmatizer <- textstem::lemmatize_words #Function that returns the lemmatized version of the input token
customlexicon <- #Vocabulary that the lemmatizer uses (by default use lexicon::hash_lemmas; modifications made below)
  lexicon::hash_lemmas %>% 
    mutate(
      lemma =
        case_when(
          token == "broker" ~ "broker", #(was "broke" lemma by default)
          str_starts(token,"renew") ~ "renewal", #(putting these all in the same lemma, was mapped to separate "renew" and "renewal" lemmas by default)
          TRUE ~ lemma
          )
    ) %>%
    add_row(token = "renew",lemma = "renewal") %>% #additional row to complete the remapping of "renew"-type tokens
    add_row(token = "info",lemma = "information") %>% #additional row to take care of common word shortening
    arrange(token)
stopwords <- stopwords::stopwords() #Vector of tokens considered stopwords
negations <- c(stringr::str_subset(stopwords::stopwords(), "^no|'t$"), "cannot", "mightn't", "needn't") #Tokens identified as 'negation' tokens
lemma_inclusion_percentile <- 0.88 #Percentile of word occurrence (at the opportunity level) for inclusion in the dictionary


#SVD parameters
topics <- 8 #Number of topics in the decomposition
what <- "fastestword" #Method for parsing text (see 'quanteda::tokens')
scheme <- "inverse" #Method for tf-idf
k <- 0.5 #Parameter for tf-idf
base <- exp(1) #Parameter for tf-idf


###IMPORT ACTIVITIES###

cat("\n\tImporting data...")

#(note: if this SVD is intended to be used in a modeling project, only import the activities from the Train records)
setup <-
  read_csv(
    file = str_c(InPath, "IllustrativeSalesActivities_Preproc.csv"),
    col_types = cols()
  ) %>%

    #rename some columns
    rename(
      ActivityID=Actv_Index,
      OpportunityID=Opp_Index,
      Text=Mod_Description
      ) %>%
  
    #select just those columns
    select(
      ActivityID
      ,OpportunityID
      ,Text
    )
  
cat("DONE\n")


###BUILD DICTIONARY###

cat("\n\tFinding unique lemmas...")

#White-space separated
all_tokens <-
  setup %>%
  
  #Send tokens down the rows by whitespace tokenization
  separate_rows(
    Text,
    sep = token_split
  ) %>%
  
  #Indicacte lemmas, stopwords, negation
  mutate(
    Lemma = lemmatizer(Text, dictionary=customlexicon)
  ) 

#Find master list of unique lemmas in dictionary
final_lemmas <-
  
  all_tokens %>%
  
  #Remove non-negated stop words
  filter(
    !((Lemma %in% stopwords) & !(Lemma %in% negations))
  ) %>%
  
  #Find word counts in each group
  group_by(
    Lemma
  ) %>%
  summarise(
    Opportunities = n_distinct(OpportunityID),
    .groups = "drop"
  ) %>%
      
  #Select top lemmas
  top_n(
    ceiling((1 - lemma_inclusion_percentile)*n_distinct(Lemma)),
    Opportunities
  ) %>%
  
  #Keep distinct values
  pull(Lemma) %>%
  unique %>% 
  sort

cat("DONE\n")

#Make a data frame and assign index to each lemma
final_lemmas <-
  tibble(
    lemma_id = seq_along(final_lemmas),
    value = final_lemmas
  )

#Make the dictionary by mapping all associated raw tokens to the lemma
#For completeness, include all rows found in lexicon for each final lemma, plus a row for when the token = lemma
final_dictionary <-
  rbind(
    final_lemmas %>%
    
      inner_join(
        y = 
          customlexicon %>%
          
          select(
            Text = token,
            value = lemma
          ),
        by = "value"
      ),
    
    final_lemmas %>%
      
      mutate(
        Text = value
      ) 
    
  ) %>%
  
  distinct %>%
  
  arrange(lemma_id,Text) %>%
  
  #Add ID
  mutate(
    token_id = seq_along(Text)
  )

#Write dictionary to file
  
cat("\n\tWriting to file...")
  
  #Write the lemma file
  final_lemmas %>%
    write_csv(
      file = str_c(OutPath, "dictionary_lemmas.csv"),
      na = ""
    )
  
  final_dictionary %>%
    
    #Remove lemma
    select(
      token_id,
      lemma_id,
      value = Text
    ) %>%
    
    #Write the index lookup for the raw tokens
    write_csv(
      file = str_c(OutPath, "dictionary_tokens.csv")
    )

#Finish dictionary
final_dictionary <-
  final_dictionary %>%
  
  #For each lemma
  split(.$value) %>%
  map(
    pluck,
    "Text"
  ) %>%
  
  #Formalize the dictionary object
  dictionary()

cat("DONE\n")


###CREATE DOCUMENT TERM MATRIX###

cat("\n\tCreating document-term matrix...")

#Concatenate text up to the opportunity level
documents <-
  setup %>%
  
  #Convert NA's to empty strings
  mutate(
    Text = 
      Text %>%
      str_replace_na(
        replacement = "NA"
      )
  ) %>%
  
  #For opportunity, collapse the text
  group_by(
    OpportunityID
  ) %>%
  summarise(
    Text = 
      str_c(
        Text,
        collapse = " "
      ),
    .groups = "drop"
  )

#Make the initial document-term matrix
dtm_initial <-
  dfm(
    tokens_lookup(
      x =
        quanteda::tokens(
          documents$Text,
          what = what
        ),
      dictionary = final_dictionary
    )
  )

#Inverse document weighting
dtm_docweights <-
  docfreq(
    x = dtm_initial,
    scheme = scheme,
    k = k,
    base = base
  )

#Get tf-idf matrix 
dtm_tfidf_adjusted <-
  dfm_weight(
    x = dtm_initial,
    weights = dtm_docweights
  )

cat("DONE\n")

cat("\n\tWriting IDF Adjustments (by Lemma) to file...\n")

#Write document weights by lemma to file
IDF_Weights <-
  dtm_docweights %>%
  
  #Make a tibble
  enframe(
    name = "Lemma",
    value = "value"
  ) %>%
  
  #Replace with index
  mutate(
    lemma_id = seq_along(Lemma)
  ) %>%
  select(
    lemma_id,
    value
  )

IDF_Weights %>%
  write_csv(
    file = str_c(OutPath, "IDF_Weights.csv")
  )

cat("DONE\n")


###SINGULAR VALUE DECOMPOSITION###

cat("\n\tRunning SVD...")

#Run SVD
svd <-
  textmodel_lsa(
    x = dtm_tfidf_adjusted,
    nd = topics,
    margin = "both"
  )

#Plot Variability Explained by Number of Topics (capped at a max of 50)
#Suppressed warning is related to requesting ALL Singular Values (rather than the first k),
#but that's okay because that's what we intend to do here to build the plot.
singularvaluesforplot <-
  suppressWarnings(
    textmodel_lsa(
      x = dtm_tfidf_adjusted,
      nd = min(nrow(dtm_tfidf_adjusted),ncol(dtm_tfidf_adjusted),50),
      margin = "both"
    )$sk
  )

plotdata <-
  data.frame(
    topic_id = seq_along(singularvaluesforplot),
    SingularValue = singularvaluesforplot,
    PctVarianceExplained = cumsum(singularvaluesforplot^2)/sum(singularvaluesforplot^2)*100,
    LTSelectedTopics = 
      case_when(
        seq_along(singularvaluesforplot) < topics ~ "Y", #less than because the plot colors line up to the first "N"
        TRUE ~ "N"
      )
  )

VarianceExplainedPlot <-
  ggplot(data=plotdata) +
  aes(x=topic_id, y=PctVarianceExplained, group=1) +
  geom_line(aes(color=LTSelectedTopics, size=LTSelectedTopics)) +
  scale_color_manual(
    breaks = c("Y","N"),
    values = c("#0070C0","#969696")
  ) +
  scale_fill_manual(
    breaks = c("Y","N"),
    values = c("#0070C0","#969696")
  ) +
  scale_size_manual(
    breaks = c("Y","N"),
    values=c("Y"=1.5, "N"=0.75)
  ) +
  ylim(0,100) +
  xlim(1,min(nrow(dtm_tfidf_adjusted),ncol(dtm_tfidf_adjusted),50)) +
  xlab("Topics") +
  ylab("% of Variance Explained") +
  labs(color="Selected Topic", size="Selected Topic") +
  ggtitle("Cumulative Percent of Variance Explained")

VarianceExplainedPlot

cat("DONE\n")

#Write document topic scores (u matrix), singular values (sigma matrix), and topic lemma weights (v matrix) to files

cat("\n\tWriting document topic scores (u matrix), singular values (sigma matrix), and topic lemma weights (v matrix) to files...\n")

##document topic scores (u matrix)
Document_Topic_Scores <-
  as.matrix(svd$docs) %>%
  
  #Convert to data frame, add id
  as.data.frame() %>%
  mutate(
    opportunity_id = 1:nrow(.)
  ) %>%
  as_tibble %>%
  
  #Send components down the rows
  gather(
    key = "topic_id",
    value = "value",
    -opportunity_id
  ) %>%

  #Remove leading character
  mutate(
    topic_id = 
      topic_id %>%
      str_remove(
        pattern = "^V"
      ) %>%
      as.numeric()
  )

Document_Topic_Scores %>%
  write_csv(
    file = str_c(OutPath, "Document_Topic_Scores.csv")
  )

##singular values (sigma matrix)
Singular_Values <- 
  tibble(
    topic_id = seq_along(svd$sk),
    value = svd$sk
  )

Singular_Values %>%
  write_csv(
    file = str_c(OutPath, "Singular_Values.csv")
  )

##topic lemma weights (v matrix)
Topic_Lemma_Weights <- 
  as.matrix(svd$features) %>% 
  
  #Convert to data frame, add rownames as a column
  as.data.frame() %>%
  mutate(
    lemma = rownames(.),
    lemma_id = 1:nrow(.)
  ) %>%
  as_tibble %>%
  
  #Send components down the rows
  gather(
    key = "topic_id",
    value = "value",
    -lemma,
    -lemma_id
  ) %>%
  
  #Remove leading character
  mutate(
    topic_id = 
      topic_id %>%
      str_remove(
        pattern = "^V"
      ) %>%
      as.numeric()
  ) %>%
  
  #Change column order
  relocate(
    topic_id
    ,lemma_id
    ,lemma
    ,value
  )

Topic_Lemma_Weights %>%
  write_csv(
    file = str_c(OutPath, "Topic_Lemma_Weights.csv")
  )

cat("DONE\n")


###DEMONSTRATE HOW WE CAN GET "TOPIC" SCORES GIVEN A TFIDF-ADJUSTED DTM###

#Multiply TFIDF-Adjusted DTM by 'v' matrix times the inverted singular values to get "topic" scores for each document 
Document_Topic_Scores_Calc <-
  as.matrix(dtm_tfidf_adjusted %*% (svd$features %*% diag(1/svd$sk))) %>%
  
  #Convert back to tibble
  as.data.frame() %>%
  as_tibble()

#Show equivalence to 'u' matrix
Document_Topic_Scores_U <-
  as.matrix(svd$docs) %>%
  
  #Convert back to tibble
  as.data.frame() %>%
  as_tibble()

cat('\n\tTesting Equality of calculated topic scores to SVD "u" matrix...',all.equal(Document_Topic_Scores_Calc,Document_Topic_Scores_U),"\n")
