#Note this program keeps the SVD component dataframes in a "long" format while doing the math to compute topic scores
#rather than in the "wide" shape of a matrix, but the result will be the same.
#(less data transformation/transposition required this way)

#The final topic scores output will still be pivoted into "wide" so the topic variables are each in their own column.

###PACKAGES###
require(dplyr)
require(tibble)
require(tidyr)
require(readr)
require(stringr)


###FILEPATHS AND PARAMETERS###

#Filepaths
InPath <- "C:/Users/UserName/Documents/SVD_Illustrative_Example/"
SVDComponentsPath <- "C:/Users/UserName/Documents/SVD_Illustrative_Example/Results/"
OutPath <- "C:/Users/UserName/Documents/SVD_Illustrative_Example/Results/"

#Dictionary filename
Dict_Filename <- "dictionary_tokens.csv"

#IDF Weights filename
IDF_Weights_Filename <- "IDF_Weights.csv"

#SVD Components filenames
Singular_Values_Filename <- "Singular_Values.csv"
Topic_Lemma_Weights_Filename <- "Topic_Lemma_Weights.csv"


###IMPORT ACTIVITIES###

cat("\n\tImporting data...")

opps_to_score <-
  read_csv(
    file = str_c(InPath, "zz_AddlIllustrativeSalesActivities_ToGiveTopicScores_Preproc.csv"),
    col_types = cols()
  ) %>%
  
  #rename some columns
  rename(
    ActivityID=Actv_Index,
    OpportunityID=Opp_Index,
    Text=Mod_Description
  ) %>%
  
  #select just those columns
  select(
    ActivityID
    ,OpportunityID
    ,Text
  )

cat("DONE\n")


###CREATE DOCUMENT TERM MATRIX###

cat("\n\tImporting dictionary and IDF weights...")

#Import Dictionary and IDF Weights
dictionary_tokens <- 
  read_csv(
    file = str_c(SVDComponentsPath, Dict_Filename),
    col_types = cols()
  )

IDF_Weights <-
  read_csv(
    file = str_c(SVDComponentsPath, IDF_Weights_Filename),
    col_types = cols()
  )

cat("DONE\n")


cat("\n\tCreating document-term matrix...")

#Concatenate text up to the opportunity level
dtm_tfidfadj_to_score <-
  opps_to_score %>%
  
  #Convert NA's to empty strings
  mutate(
    Text = 
      Text %>%
      str_replace_na(
        replacement = "NA"
      )
  ) %>%
  
  #Send tokens down the rows
  separate_rows(
    Text,
    sep = " "
  ) %>%
  
  #Join to get lemma_id for the token
  inner_join(
    y = dictionary_tokens,
    by = 
      c(
        "Text" = "value"
      )
  ) %>%
  
  #Compute the occurrence for each lemma
  group_by(
    OpportunityID,
    lemma_id
  ) %>%
  summarise(
    N = n(),
    .groups = "drop"
  ) %>%
  
  #Join to get document weight for each lemma
  inner_join(
    y = IDF_Weights,
    by = 
      c(
        "lemma_id"
      )
  ) %>%

  #Compute tf-idf
  mutate(
    N = N * value
  ) %>%
    select(
      -value
    )

cat("DONE\n")


###COMPUTE TOPIC SCORES###

cat("\n\tImporting singular values and topic lemma weights...")

#Import singular values and topic lemma weights
Singular_Values <- 
  read_csv(
    file = str_c(SVDComponentsPath, Singular_Values_Filename),
    col_types = cols()
  )

Topic_Lemma_Weights <-
  read_csv(
    file = str_c(SVDComponentsPath, Topic_Lemma_Weights_Filename),
    col_types = cols()
  )

cat("DONE\n")


cat("\n\tComputing topic scores...")

topic_scores <-
  dtm_tfidfadj_to_score %>%
  
  #Join to get [Topic Lemma Weight * 1/Singular_Value]
  inner_join(
    y =
      Topic_Lemma_Weights %>%
      
      #Join to get singular values for the appropriate index
      inner_join(
        y = 
          Singular_Values %>%
          
          #Rename column
          rename(
            singular_value = value
          ),
        by =
          c(
            "topic_id"
          )
      ) %>%
      
      #Compute the value
      mutate(
        value = value/singular_value
      ) %>%
      
      #Remove uneeded column
      select(
        -singular_value
      ),
    by = 
      c(
        "lemma_id"
      )
  ) %>%
  
  #Compute weighted-sum within each opportunity, topic
  group_by(
    OpportunityID,
    topic_id
  ) %>%
  summarise(
    value = sum(N * value),
    .groups = "drop"
  ) %>%
  
  #Spread the SVD topic scores across the columns
  pivot_wider(
    names_from = topic_id,
    values_from = value,
    names_prefix = "V"
  ) %>%
    
  #Fill in missing scores with 0's
  mutate(
    across(
      starts_with("V"),
      coalesce,
      0
    )
  ) 

cat("DONE\n")


cat("\n\tWriting document topic score variables to file...\n")

topic_scores %>%
  write_csv(
    file = str_c(OutPath, "zz_TopicScoringNewDocs_Topic_Scores.csv")
  )

cat("DONE\n")